#include "RTreeC.h"

#include <octave/parse.h>

#include <unistd.h>

RTreeC::RTreeC()
{
  error_with_id("RTree:BUG", "default constructor called");
}

RTreeC::RTreeC(rtree_t *rtree)
{
  this->rtree = rtree;
}

RTreeC::RTreeC(std::size_t dim,
               std::string split,
               std::uint32_t node_page)
{
  state_flags_t flags = state_flags(split, node_page);

  if ((this->rtree = rtree_alloc()) == NULL)
    error_with_id("RTree:memory", "failed rtree_alloc()");
  else
    {
      if (rtree_init(this->rtree, dim, flags) != 0)
        {
          rtree_destroy(this->rtree);
          error_with_id("RTree:ERRNO", "%s", strerror(errno));
        }
    }
}

RTreeC::~RTreeC(void)
{
  rtree_destroy(this->rtree);
}

// static (class) methods

RTreeC* RTreeC::bsrt_read(int fd)
{
  int fd2;

  if ((fd2 = dup(fd)) == -1)
    error_with_id("RTree:ERRNO", "dup: %s", strerror(errno));

  FILE *st;

  if ((st = fdopen(fd2, "rb")) == NULL)
    error_with_id("RTree:ERRNO", "fdopen: %s", strerror(errno));

  errno = 0;

  rtree_t *rtree = rtree_bsrt_read(st);

  fclose(st);

  if (rtree == NULL)
    {
      if (errno)
        error_with_id("RTree:ERRNO", "%s", strerror(errno));
      else
        error_with_id("RTree:librtree", "parse of BSRT");
    }

  return new RTreeC(rtree);
}

RTreeC* RTreeC::csv_read(int fd,
                         std::size_t dim,
                         std::string split,
                         std::uint32_t node_page)
{
  state_flags_t flags = state_flags(split, node_page);

  int fd2;

  if ((fd2 = dup(fd)) == -1)
    error_with_id("RTree:ERRNO", "dup: %s", strerror(errno));

  FILE *st;

  if ((st = fdopen(fd2, "r")) == NULL)
    error_with_id("RTree:ERRNO", "fdopen: %s", strerror(errno));

  errno = 0;

  rtree_t *rtree = rtree_csv_read(st, dim, flags);

  fclose(st);

  if (rtree == NULL)
    {
      if (errno)
        error_with_id("RTree:ERRNO", "%s", strerror(errno));
      else
        error_with_id("RTree:librtree", "parse of CSV");
    }

  return new RTreeC(rtree);
}

RTreeC* RTreeC::json_read(int fd)
{
  int fd2;

  if ((fd2 = dup(fd)) == -1)
    error_with_id("RTree:ERRNO", "dup: %s", strerror(errno));

  FILE *st;

  if ((st = fdopen(fd2, "r")) == NULL)
    error_with_id("RTree:ERRNO", "fdopen: %s", strerror(errno));

  errno = 0;

  rtree_t *rtree = rtree_json_read(st);

  fclose(st);

  if (rtree == NULL)
    {
      if (errno)
        error_with_id("RTree:ERRNO", "%s", strerror(errno));
      else
        error_with_id("RTree:librtree", "parse of BSRT");
    }

  return new RTreeC(rtree);
}

// static (class) constants

const std::string RTreeC::name = rtree_package_name;
const std::string RTreeC::bugreport = rtree_package_bugreport;
const std::string RTreeC::url = rtree_package_url;
const std::string RTreeC::version = rtree_package_version;

// instance methods

void RTreeC::add_rect(rtree_id_t id, std::vector<rtree_coord_t> rect)
{
  validate_rect(rect);

  int err = rtree_add_rect(this->rtree, id, rect.data());

  if (err != 0)
    error_with_id("RTree:librtree", "%s", rtree_strerror(err));
}

std::size_t RTreeC::branching_factor(void)
{
  return rtree_branching_factor(this->rtree);
}

std::size_t RTreeC::branch_size(void)
{
  return rtree_branch_size(this->rtree);
}

void RTreeC::bsrt_write(int fd)
{
  int fd2 = dup(fd);

  if (fd2 == -1)
    error_with_id("RTree:ERRNO", "dup: %s", strerror(errno));

  FILE *st;

  if ((st = fdopen(fd2, "wb")) == NULL)
    error_with_id("RTree:ERRNO", "fdopen: %s", strerror(errno));

  int err = rtree_bsrt_write(this->rtree, st);

  fclose(st);

  if (err != 0)
    error_with_id("RTree:librtree", "%s", rtree_strerror(err));
}

std::size_t RTreeC::bytes(void)
{
  return rtree_bytes(this->rtree);
}

RTreeC* RTreeC::clone(void)
{
  return new RTreeC(rtree_clone(this->rtree));
}

std::size_t RTreeC::dim(void)
{
  return rtree_dims(this->rtree);
}

bool RTreeC::empty(void)
{
  return rtree_empty(this->rtree);
}

std::uint16_t RTreeC::height(void)
{
  return rtree_height(this->rtree);
}

bool RTreeC::identical(std::shared_ptr<RTreeC> other)
{
  return rtree_identical(this->rtree, other->rtree);
}

void RTreeC::json_write(int fd)
{
  int fd2 = dup(fd);

  if (fd2 == -1)
    error_with_id("RTree:ERRNO", "dup: %s", strerror(errno));

  FILE *st;

  if ((st = fdopen(fd2, "w")) == NULL)
    error_with_id("RTree:ERRNO", "fdopen: %s", strerror(errno));

  int err = rtree_json_write(this->rtree, st);

  fclose(st);

  if (err != 0)
    error_with_id("RTree:librtree", "%s", rtree_strerror(err));
}

std::size_t RTreeC::node_size(void)
{
  return rtree_node_size(this->rtree);
}

std::size_t RTreeC::page_size(void)
{
  return rtree_page_size(this->rtree);
}

void RTreeC::postscript(int fd,
                        std::shared_ptr<RTreeStyleC> rtsc,
                        std::string axis_name,
                        float extent,
                        float margin)
{
  extent_axis_t axis;

  if (axis_name == "width")
    axis = axis_width;
  else if (axis_name == "height")
    axis = axis_height;
  else
    error_with_id("RTree:args-invalid",
                  "bad AXIS %s", axis_name.c_str());

  rtree_postscript_t ps = {
    .style = rtsc->style,
    .axis = axis,
    .extent = extent,
    .margin = margin,
    .title = "librtree-octave output"
  };

  int fd2 = dup(fd);

  if (fd2 == -1)
    error_with_id("RTree:ERRNO", "dup: %s", strerror(errno));

  FILE *st;

  if ((st = fdopen(fd2, "w")) == NULL)
    error_with_id("RTree:ERRNO", "fdopen: %s", strerror(errno));

  int err = rtree_postscript(this->rtree, &ps, st);

  fclose(st);

  if (err != 0)
    error_with_id("RTree:librtree", "%s", rtree_strerror(err));
}

std::size_t RTreeC::rect_size(void)
{
  return rtree_rect_size(this->rtree);
}

int RTreeC::search(std::vector<rtree_coord_t> rect, search_context_t *context)
{
  validate_rect(rect);
  return rtree_search(this->rtree, rect.data(), search_cb, context);
}

int RTreeC::update(update_context_t *context)
{
  return rtree_update(this->rtree, update_cb, context);
}

double RTreeC::unit_sphere_volume(void)
{
  return rtree_unit_sphere_volume(this->rtree);
}

// private methods

state_flags_t RTreeC::state_flags(std::string split, std::uint32_t node_page)
{
  state_flags_t split_flag;

  if (split == "quadratic")
    split_flag = RTREE_SPLIT_QUADRATIC;
  else if (split == "linear")
    split_flag = RTREE_SPLIT_LINEAR;
  else if (split == "greene")
    split_flag = RTREE_SPLIT_GREENE;
  else
    error_with_id("RTree:args-invalid",
                  "unknown SPLIT: %s",
                  split.c_str());

  state_flags_t
    node_page_flag = RTREE_NODE_PAGE((state_flags_t)node_page),
    flags = split_flag | node_page_flag;

  return flags;
}

void RTreeC::validate_rect(std::vector<rtree_coord_t> rect)
{
  std::size_t
    dims = rtree_dims(this->rtree),
    size = rect.size();

  if (size != 2 * dims)
    error_with_id("RTree:args-invalid",
                  "rectangle array size %zu, should be %zu",
                  size, 2 * dims);
}

// callbacks

static int update_cb(rtree_id_t id, rtree_coord_t *rect, void *varg)
{
  update_context_t *arg = static_cast<update_context_t*>(varg);
  octave::interpreter *interp = arg->interp;
  octave_value f = arg->f, context = arg->context;
  size_t dim = arg->dim;
  ColumnVector cv_rect_in(2 * dim);

  {
    double *data = cv_rect_in.fortran_vec();
    for (std::size_t i = 0 ; i < 2 * dim ; i++)
      data[i] = rect[i];
  }

  octave_value_list cb_args;

  cb_args(0) = octave_uint64(id);
  cb_args(1) = cv_rect_in;
  cb_args(2) = context;

  // in the case that the user supplied function does not assign
  // the output variable, raises 'Octave:undefined-function', see
  // the tests -- perhaps catch those and return 1, later

  octave_value_list retval = interp->feval(f, cb_args, 1);

  // in the case that retval(0) cannot be converted to an N-D array
  // raises with message 'invalid conversion ...'

  ColumnVector cv_rect_out = retval(0).column_vector_value();

  if (cv_rect_out.dims().numel() != 2 * dim)
    error_with_id("RTree:args-number",
                  "callback returned a %lu-array, expected %lu",
                  cv_rect_out.dims().numel(), 2 * dim);

  {
    double *data = cv_rect_out.fortran_vec();
    for (std::size_t i = 0 ; i < 2 * dim ; i++)
      rect[i] = data[i];
  }

  return 0;
}

// search callback, rather similar to update, but the callback
// context must be returned, and that is then assigned to the
// argument (so it is incrementally modified).

static int search_cb(rtree_id_t id, void *varg)
{
  search_context_t *arg = static_cast<search_context_t*>(varg);
  octave::interpreter *interp = arg->interp;
  octave_value f = arg->f, context = arg->context;
  octave_value_list cb_args;

  cb_args(0) = octave_uint64(id);
  cb_args(1) = context;

  // in the case that the user supplied function does not assign
  // the output variable, raises 'Octave:undefined-function', see
  // the tests -- perhaps catch those and return 1, later

  octave_value_list retval = interp->feval(f, cb_args, 1);

  // retval(0) should be the modified context, then we update

  arg->context = retval(0);

  return 0;
}
